classdef Agent <hgsetget
    
    
    properties
        face;       % Its interface for plotting, displaying and saving/importing (class).
        nn;         % Its mental model/neural network (class).
        kn;         % Its mental model/neural network of knowledge (class).
        lm;         % Its learning model (reinforcement learning (class).
        policy;     % Its  policy (class).
        experiences;% The agents stored experiences (class).
        et;         % The agents eligibility trace (class).
        numTrainings; % Tracks how often the agent has been trained (does not count retraining in eligibility traces).
        useRL;      % Set to use/not use Reinforcement learning
        useSprimByAgent; %
        sPrimMemoryLength; %
        numActions; % Number of actions to choose from. Can only be one.
        numStates;  % Number of states, can be one or two.
        numInputs;  % Number of actions + number of states.
        sPrimType;  % Complexity understanding, 1 s, 2 s-sa, 3, s-sa, s-as-s^2
    end
    
    properties (SetAccess='private')
        sPrimParams; % Computed constants for predicting the next state
        N;  % used for calculating mental model
        H;  % used for calculating mental model
        X;  % used for calculating mental model
        Y;  % used for calculating mental model
        AC; % actions to test mental model
        ST; % states to test mental model
        MMtrace;
        checkEtraceEffect;
        sPrimValue;
    end
    
    methods
        
        function obj = Agent (varargin)
            file             = varargin{1};
            obj.face         = Face(obj,file);
            
            for i=1:(length(varargin)-1)/2
                obj.(varargin{i*2})(varargin{i*2+1})
            end
            obj.numTrainings    = 0;
            obj.useSprimByAgent = 1;
            obj.sPrimParams      = [1;0;0;0];
            obj.sPrimMemoryLength = 30; % -1 means full memory
            obj.N = 20;
            obj.H = linspace(0,1,obj.N);
            [obj.X,obj.Y] = meshgrid(obj.H,obj.H);
            obj.AC = reshape(obj.X,1,obj.N*obj.N); % actions to test mental model
            obj.ST = reshape(obj.Y,1,obj.N*obj.N); % states to test mental model
            obj.checkEtraceEffect = false;
            obj.sPrimType = 2;
            obj.sPrimValue = 0;
        end
        
        function stopRun=train(obj, input, reward) % input = aPrim and sPrim
            cutTrace = false;
            accCutTrace = false;
            stopRun = obj.trainCT(input, reward, cutTrace, accCutTrace);
        end
        function stopRun=trainCT(obj, input, reward, cutTrace, accCutTrace) % input = aPrim and sPrim
            
            Qnew       = -1; % need if no reinforcement learning
            stopRun    = false; % break if Q value is too large
            aPrim      = input(obj.numActions:obj.numInputs-obj.numStates,:);
            sPrim      = input(obj.numActions+1:obj.numInputs,:);
            
            if obj.numTrainings == 0
                Qnew = 0;%reward*0.01; % cant be to close cause it wont train next time then?
                obj.numTrainings = obj.numTrainings+1;
                sPrimAgent = sPrim;
                obj.experiences.add (aPrim, sPrim, reward, Qnew, obj.numTrainings, sPrimAgent, aPrim);
                
                
            else
                a=obj.experiences.getLatestAction();
                s=obj.experiences.getLatestState();
                r=obj.experiences.getLatestReward();
                sPrimAgent = obj.sPrimValue;% has been calculated whe ngetPolicy was called obj.calcSprimAgent(a, s);
                
                % ----------  No RL No traces  -----------------------
                if obj.useRL == false
                    obj.nn.train([a; s'], r);
                    obj.numTrainings = obj.numTrainings + 1;
                    obj.experiences.add (aPrim, sPrim, reward, Qnew, obj.numTrainings, sPrimAgent, obj.policy.greedy);
                    
                    % -----------  RL  -----------------------------------
                elseif obj.useRL == true
                    
                    if obj.useSprimByAgent
                        omega = obj.lm.calcTDerror (aPrim, sPrimAgent, r, a, s); % the current error
                    else
                        omega = obj.lm.calcTDerror (aPrim, sPrim, r, a, s); % the current error
                    end
                    if omega == -Inf % disp('reward signal too high, LR possibly too high');
                        stopRun=true;
                    end
                    Qnew = obj.lm.calcQnew(a, s, omega); % new Q with current error
                    obj.nn.train([a; s'], Qnew);
                    %obj.nn.LW
                    obj.numTrainings = obj.numTrainings + 1;
                    obj.experiences.add (aPrim, sPrim, reward, Qnew, obj.numTrainings, sPrimAgent, obj.policy.greedy);
                    
                    % ---------  Traces  ----------------------------
                    if obj.lm.useEt == true
                        obj.et.add(a, s, Qnew, obj.numTrainings, obj.experiences); % add new a,s, and trace = 1
                        etSize  = obj.et.getSize(); % loop the Etrace vector
                        
                        if cutTrace == true && accCutTrace == true && etSize > 0
                            obj.et.clearTrace(obj.experiences);
                        else
                            j = 0; % counter for experiences vector
                            adjBias = obj.nn.getAdjBias();
                            obj.nn.setadjBias(0);
                            obj.MMtrace(:, :)=[];
                            
                            for i = (etSize - 1):-1:1 % i counter for trace vector
                                trace       = obj.et.getTrace(i); % current trace
                                retrainPos  = obj.numTrainings - j - 1; % get next last position
                                
                                if retrainPos > 0 && trace > 0.05 % to avoid uneccessary training
                                    aTr       = obj.experiences.getAction(retrainPos); % action to train/retrain
                                    sTr       = obj.experiences.getState(retrainPos); % state to train/retrain
                                    QnewTr    = obj.lm.calcQnewTrace(aTr, sTr, omega, trace); % newQ to train/retrain with
                                    
                                    obj.nn.train ([aTr; sTr'], QnewTr); % first time its regular train then its retraining
                                    obj.et.updateQAccDiff(i, QnewTr, retrainPos);
                                    % check new mental model
                                    if obj.checkEtraceEffect == true
                                        for q=1:length(obj.AC)
                                            obj.MMtrace(j+1,q) = obj.sim([obj.AC(q); obj.ST(q)]);
                                        end
                                    end
                                    j = j + 1;
                                else
                                    accDiff = obj.et.getAccDiff(i);
                                    % add trace to experiences vector (now that the trace wont be updated anymore)
                                    obj.experiences.updateQaccDiff(retrainPos, accDiff);
                                    break;
                                end
                            end
                            obj.nn.setadjBias(adjBias);
                            obj.et.discountTrace();
                        end
                    end
                end
                %disp(['total num trainings ' num2str(obj.numTrainings)])
            end
        end
        
        % agent uses the latest real states 
        function out = calcSprimAgent(obj, a, s)
            if obj.numTrainings > 5
                if obj.sPrimMemoryLength == -1
                    memoryLength = length(obj.experiences.states);
                else
                    memoryLength = obj.sPrimMemoryLength;
                end
                memLen = min(memoryLength, length(obj.experiences.states));
                memStart = max(1, length(obj.experiences.states) - memLen );
                
                Y1       = obj.experiences.states(memStart:length(obj.experiences.states));
                Y1(1)    = [];
                
                %or ?
                %Y(length(Y),:) = [];
                statesMemory=obj.experiences.states(memStart:length(obj.experiences.states));
                actionsMemory=obj.experiences.actions(memStart:length(obj.experiences.actions));

                if obj.sPrimType == 1
                    X1                  = [statesMemory];
                    X1(length(X1),:)    = [];
                    XY1                 = X1\Y1;
                    obj.sPrimParams     = XY1;
                    sPrimAgent          = obj.sPrimParams(1)*s;
                elseif obj.sPrimType == 2
                    X2  = [statesMemory, statesMemory.*actionsMemory'];
                    X2(length(X2),:)    = [];
                    try 
                        XY2                 = mldivide(X2,Y1);
                        warning('off','last');
                        %error(lastwarn); % makes warning into error, hapens whe nX and Y are zero
                    catch the_warning
                        %X2
                        %Y1
                        %XY2
                        %sPrimAgent          = obj.sPrimParams(1)*s + s*a*obj.sPrimParams(2)
                        %disp('catch me!')
                        %warning 
                        %XY2
                    end
                    obj.sPrimParams     = XY2;
                    sPrimAgent          = obj.sPrimParams(1)*s + s*a*obj.sPrimParams(2);
                elseif obj.sPrimType == 3
                    X3  = [statesMemory, statesMemory.*actionsMemory', statesMemory.^2];
                    X3(length(X3),:)    = [];
                   
                    XY3                 = X3\Y1;
                    obj.sPrimParams     = XY3;
                    sPrimAgent          = obj.sPrimParams(1)*s + ...
                        s*a*obj.sPrimParams(2) + ...
                        s.^2*obj.sPrimParams(3);
                elseif obj.sPrimType == 4
                    X4  = [statesMemory, statesMemory.*actionsMemory', statesMemory.^2, ...
                        statesMemory.^3];
                    X4(length(X4),:)    = [];
                    XY4                 = X4\Y1;
                    obj.sPrimParams     = XY4;
                    sPrimAgent          = obj.sPrimParams(1)*s + ...
                        s*a*obj.sPrimParams(2) + ...
                        s.^2*obj.sPrimParams(3) + ...
                        s.^3*obj.sPrimParams(4);
                else
                    disp('error in AgentX:calcSprimAgent')
                end
                out = max(0, sPrimAgent);
                K = 1;% carrying capacity (take max s?)
                out = min(out, 2*K);
                obj.sPrimValue = out;
            else
                obj.sPrimParams=[1;0;0;0];
                out = max(0, s);
                obj.sPrimValue = out;
            end
        end
        
        
        function output = sim(obj, input)
            output=obj.nn.sim(input);
        end
        
        function output = getMentalModel(obj, input)
            disp('use sim instead?')
            pause
            for q=1:length(input)
                output(q)=obj.nn.sim(input(:,q));
            end
        end
        
        function output = getNN(obj)
            output=obj.nn;
        end
        
        function setNN(obj, new_nn)
            obj.nn = new_nn;
        end
        
        function output = getNNLW(obj)
            output=obj.nn.LW;
        end
        
        % CHOOSE ACTIONS
        function action = getBestAction(obj, state)
            action=obj.policy.calcBestPolicy(state);
        end
        
        function action = getPolicy(obj, state)
            action = obj.getSoftmaxPolicy(state);
        end
        
        function action = getSoftmaxPolicy(obj, state)
            action = obj.policy.calcSoftmaxPolicy(state);
        end
        
        % must only be called after getSoftmaxPolicy
        function level = getExperimentationSoftmaxVSGreedy(obj)
            level = obj.policy.getExpSMvsGreedy();
        end
        
        function omega = getOmega(obj)
            omega = obj.lm.getOmega();
        end
        
        
        % GETTERS
        function experiences = getExperiences(obj)
            experiences = obj.experiences;
        end
        
        % row 1: a, 2: s, 3: r, 4: Qnew, 5: Qacc, 6: trainig no, nt
        function out = getExperiencesVector(obj)
            out = obj.experiences.getExperiences();
        end
        
        
        function learningrate = getLearningRate(obj)
            learningrate = obj.nn.lr;
        end
        % SETTERS
        function setRandValues(obj, useSameRandSeq, value)
            obj.policy.setRandValues(useSameRandSeq, value);
        end
        
        function toggleRandom(obj, value)
            obj.setRandValues(true, value)
        end
        
        function setTemperature(obj,temp)
            obj.policy.setTemperature(temp);
        end
        
        function setDiscountRate(obj, dr)
            obj.lm.setDiscountRate(dr);
        end
        
        function setStepSize(obj, ss)
            obj.lm.setStepSize(ss);
        end
        
        
        function setEpochsAndGoal(obj, numEpochs, goal)
            obj.nn.setEpochsAndGoal(numEpochs, goal);
        end
        
        function useReinforcementLearning(obj, bool)
            obj.useRL = bool;
            
        end
        
        function useEligabilityTrace(obj, bool)
            obj.lm.setUseEt(bool);
        end
        
        function setNeuronWidth(obj, value)
            obj.nn.setSpread(value);
        end
        function setLearningRate(obj, lr)
            obj.nn.setLearningRate(lr);
        end
        function setLambda(obj, value)
            obj.et.setLambda(value);
        end
        
        function setUseSprimByAgent(obj, bool)
            obj.useSprimByAgent = bool;
        end
        
        function setFaceParams(obj, learningRate,discountRate, stepSize,lambda,temperature, spread,numRbfs, useRL, useET)
            obj.face.setParams(learningRate,discountRate, stepSize,lambda,temperature, spread,numRbfs, useRL, useET);
        end
        
        function output = getSprim(obj,S,a, sprimtype)
            output = obj.getSprim(S,a,sprimtype);
            %output=S*obj.sPrimParams(1)+S*a*obj.sPrimParams(2)+S.^2*obj.sPrimParams(3);
        end
        
        function output = getSprimType(obj)
            output=obj.sPrimType;
        end
        
        % sprim 1,2,3,4 (a, as, s^2, s^3)
        function setSPrimType(obj, sPrimType)
            obj.sPrimType = sPrimType;
        end
        
        function setSPrimMemoryLength(obj, sPrimMemoryLength)
            obj.sPrimMemoryLength = sPrimMemoryLength;
        end
        
        function setCheckEtraceEffect(obj, boolean)
            obj.checkEtraceEffect = boolean;
        end
        
        function out = getMMtrace(obj)
            out = obj.MMtrace;
        end
        
        function setStartActions(obj, value)
            obj.policy.startActions = value;
        end
      
    end % end methods
end % end class


