classdef EligibilityTrace <hgsetget
    %Experiences stores the experiences of the agent
    properties
        currentSize;
        lambda;
        traceVector;
        traceVecHistory;
        tvhCtr;
        ctr2;
        maxMemLen;
        
        ikeyNt  = 1;
        iTrace  = 2;
        iQprev  = 3;
        iQaccD  = 4;
        iQorig  = 5;
 
        ia      = 6;
        is1     = 7;
        is2     = 8;
    end
    
    methods
        function obj = EligibilityTrace()
            obj.traceVector = [];            
            obj.maxMemLen   = 50;
            obj.traceVecHistory = zeros(4+obj.maxMemLen, 1);
            obj.currentSize = 0;
            obj.lambda      = 0.2;
            
        end
        
        % add a new a,s, pair
        function add(obj, a, s, Qoriginal, numTrainings, experiences)
            obj.currentSize = obj.currentSize + 1;  
            obj.traceVector(obj.ikeyNt, obj.currentSize) = numTrainings;
            obj.traceVector(obj.iQprev, obj.currentSize) = Qoriginal;
            obj.traceVector(obj.iQaccD, obj.currentSize) = 0;
            obj.traceVector(obj.iQorig, obj.currentSize) = Qoriginal;            
            obj.traceVector(obj.iTrace, obj.currentSize) = 1;
            obj.traceVector(obj.ia,     obj.currentSize) = a;
            
            if max(size(s)) == 1
                obj.traceVector(obj.is1,obj.currentSize) = s(1,1); 
            elseif max(size(s)) == 2
                obj.traceVector(obj.is1,obj.currentSize) = s(1,1); 
                obj.traceVector(obj.is2,obj.currentSize) = s(1,2);  
            end
            
            if obj.currentSize > obj.maxMemLen
                experiences.updateQaccDiff(obj.traceVector(obj.ikeyNt, obj.currentSize), ...
                    obj.traceVector(obj.iQaccD,obj.currentSize))
                obj.traceVector(:, 1) = []; % oldest trace removed
                obj.currentSize       = obj.currentSize - 1; % adjust size
            end
               % ----- trace vec history
            obj.traceVecHistory(1:4+obj.maxMemLen,numTrainings) = zeros(4+obj.maxMemLen, 1);
            ncols = size(obj.traceVecHistory, 2);
            obj.traceVecHistory(1:5, ncols) = [numTrainings a s 5 Qoriginal]';
            % ---------- end  ----
        end
        
        function updateQAccDiff(obj, i, QnewTr, numTrainings)
            diff    =  QnewTr - obj.traceVector(obj.iQprev, i);
            % set new latest Q
            obj.traceVector(obj.iQprev, i) = QnewTr;
            obj.traceVector(obj.iQaccD, i) = obj.traceVector(obj.iQaccD, i) + diff;
            
            % ---- trace vec history
            
            myCtr = obj.getTvhCtr(numTrainings) + 1;
            obj.traceVecHistory(myCtr, numTrainings) = QnewTr;
            obj.setTvhCtr(numTrainings,myCtr);

            
        end

        function out=getTvhCtr(obj, i)
            out = obj.traceVecHistory(4,i);
        end
        
        function setTvhCtr(obj, i, ctrValue)
            obj.traceVecHistory(4,i) = ctrValue;
        end
        
        function clearTrace(obj, experiences)
            i = size(obj.traceVector, 2);
            while i > 0
                experiences.updateQaccDiff(obj.traceVector(obj.ikeyNt, i), ...
                    obj.traceVector(obj.iQaccD, i))
                obj.traceVector(:, i) = [];  % remove last position
                i = size(obj.traceVector,2); % set i to new size
            end            
            obj.currentSize = 0;
        end
        
        function sizet=getSize(obj)
            sizeTV = size(obj.traceVector);
            sizet = sizeTV(1,2);    
        end
       
        function trace = getTrace(obj, index)
            trace = obj.traceVector(obj.iTrace,index);
        end
        
        function action = getAction(obj, index)
            action = obj.traceVector(obj.ia,index);
        end
        
        function state = getState(obj, index)
            if obj.agent.numInputs == 2
                state = obj.traceVector(obj.is1,index);
            elseif obj.agent.numInputs == 3
                state(1,1) = obj.traceVector(obj.is1,index);
                state(2,1) = obj.traceVector(obj.is2,index);
            end
        end
        
        function out = getQdiff(obj, index)
            out = obj.traceVector(obj.iQprev,index);
        end    
        
        function out = getAccDiff(obj, index)
            out = obj.traceVector(obj.iQaccD,index);
        end
        
        function out = getQoriginal(obj, index)
            out = obj.traceVector(obj.iQorig,index);
        end
        
        function out = getKey(obj, index)
            out = obj.traceVector(obj.ikeyNt,index);
        end
        
        function discountTrace(obj)
            traces = obj.traceVector(obj.iTrace,:).*obj.lambda; 
            obj.traceVector(obj.iTrace,:) = traces(1,:);
        end
        
        function setLambda(obj, lambda)
            obj.lambda = lambda;
        end
        
    end % end methods
end % end class


