classdef Experiences <hgsetget
    %Experiences stores the experiences of the agent
    
    properties
        actions;
        states;         % Actual (predicted) state
        sPrimByAgent;   % Predicted state by the agent
        reward;         % The actual reward from the ecosystem
        Qnew;           % The prediction value which the network is trained with
        QAccDiff;       % The accumulated difference in prediction value when retraining with traces
        trainingNo;     % 
        nt=0;           %
        greedyp;        % if greedy policy was to be chosen
    end

    methods
        function obj = Experience(obj)
            obj.trainingNo  = 0;
        end
        
        function add(obj,action, state, reward, Qnew, trainingNo, sPrimByAgent, greedyp)
            obj.actions(trainingNo)      = action;
            obj.states(trainingNo,:)     = state;
            obj.reward(trainingNo)       = reward;
            obj.Qnew(trainingNo)         = Qnew;
            obj.trainingNo(trainingNo)   = trainingNo;
            obj.nt                       = obj.nt + 1;
            obj.QAccDiff(trainingNo)     = 0;
            obj.sPrimByAgent(trainingNo) = sPrimByAgent;
            if isempty (greedyp)
                if trainingNo < 4
                    greedyp = action;
                else
                    greedyp = 999;
                end
            end
            
                
            obj.greedyp(trainingNo)      = greedyp;
        end
        
        function updateQaccDiff(obj, pos, QaccDiff)
                obj.QAccDiff(pos) = QaccDiff;
        end
        
        function updateSPrimsByAgent(obj, pos, sPrimByAgent)
            obj.sPrimByAgent(pos) = sPrimByAgent;
        end
        
        function out = getSPrimByAgent(obj, pos)
            out = obj.sPrimByAgent(pos);
        end
        
        function out=getQAccDiff(obj)
            out=obj.QAccDiff(obj.nt);
        end
        function a=getLatestAction(obj)
            a=obj.actions(obj.nt);
        end
        
        function s=getLatestState(obj)
            s=obj.states(obj.nt,:);
        end
       
         function a=getAction(obj, index)
            a=obj.actions(index);
        end
        
        function s=getState(obj, index)
            s=obj.states(index,:);
        end
        
        function r=getLatestReward(obj)
            r=obj.reward(length(obj.reward));
        end
        
        function len=getLength(obj)
            len=length(obj.trainingNo);
        end
        
         function out=getExperiences(obj)
            % a, s, r, Qnew, Qacc, trainig no, nt
            out(1,:) = obj.actions(:,:);
            out(2,:) = obj.states(:,:);
            out(3,:) = obj.reward(:,:);
            out(4,:) = obj.Qnew(:,:);
            out(5,:) = obj.QAccDiff(:,:);
            out(6,:) = obj.trainingNo(:,:);
            out(7,:) = obj.greedyp(:,:);
        end
        
    end % end methods
end % end class


