%Policy Model

classdef PolicyModel <hgsetget

    properties
        temperature; % Decides exploration level, must be between 0 and 1. 1 means complete exploration 0 meand no exploration.
        numTests = 100; % The resolution of possible actions in the Mental Model.
        numT = 30;
        possibleActions; % The vector of possible actions, size determined by numTests.
        startActions;
        expSMvsGreedy;
        greedy
    end
    
    properties (SetAccess='private')
        agent;
    end
    
    methods
        function obj=PolicyModel(agent)
            obj.temperature = 0.001;
            obj.agent       = agent;
            
            sa1 = rand*0.7 + 0.15;
            sa2 = rand*0.7 + 0.15;
            sa3 = rand*0.7 + 0.15;
            obj.startActions = [sa1 sa2 sa3];
            obj.expSMvsGreedy = 0;
        end
        
        function action=calcBestPolicy(obj, s)
            
            obj.possibleActions = linspace(0,1,obj.numT);
            % possible actions on the current state
            theState    = repmat(s,1,obj.numT); % TODO FIX REPMAT
            simInput    = [obj.possibleActions; theState];
            results     = obj.agent.sim(simInput);
            [val, index]  = max(results);
            action      = obj.possibleActions(index);
        end
        
        % probMatrix what to calcuate the softmax value of, ex. Qvalues
        function out=calcSoftmaxPolicy(obj, state)
            
            obj.possibleActions = linspace(0,1,obj.numT);
            if obj.agent.numTrainings == 0 || obj.agent.numTrainings == 1 || obj.agent.numTrainings == 2
                out = obj.startActions(obj.agent.numTrainings + 1);
                greedyp = out;
            else
                states = repmat(state, 1, obj.numT); % TODO FIX REPMAT
                input = [obj.possibleActions; states];
                Qs = obj.agent.sim(input);
                % here must have range above zero before normalizing
                min_value = min(Qs);
                if min_value < 0
                    Qs=Qs-min_value;
                end
                if max(Qs)>min(Qs)
                    Qs=Qs./max(Qs);
                    Qs=Qs.^(1/obj.temperature);
                    Qs=Qs./sum(Qs);
                    therand=rand();
                    out=obj.possibleActions(find(cumsum(Qs)>therand, 1 ));
                else
                   % disp('out is rand')
                    out=rand();
                end
                [~, index]  = max(Qs);
                greedyp      = obj.possibleActions(index);
                %if isnan( greedyp ) > 0
                    %disp('greedyp:')
                    %greedyp
                    %max(Qs)
                %end
            end
            obj.greedy = greedyp;
           
            obj.expSMvsGreedy = out-greedyp;
            
        end
            
            % Normalizes a vector by its absolute sum
        function columnVector=normabs(columnVector)
            colSum = sum(columnVector);
            columnVector = columnVector./colSum;
        end
        
        function setTemperature(obj, temp)
            obj.temperature = temp;
        end
        
        function out = getExpSMvsGreedy(obj)
            out = obj.expSMvsGreedy;
        end
        
    end % end methods
end % end class


