%RBF Neural Network

classdef RBFNeuralNetwork <hgsetget
    %RBFNeuralNetwork Creates a RBFNeuralNetwork of RBF type

    properties
        IW
        LW
        biasOutput
        numRbfs
        numInputs
        lr
        maxEpochs = 1; % num epochs to train with the same training data
        goal      = 0; % error goal - if the error is lower than this the training st
        spread;
    end
    properties (SetAccess='private')
        rbfOutput
        target
        span
        inputSpans
        Ntrains
        adjBias
    end

    methods
        function obj = RBFNeuralNetwork(numRbfs, inputSpans, lr, spread)
            obj.spread      = spread; % bias1
            obj.lr          = lr;
            obj.numRbfs     = numRbfs;
            obj.adjBias     = 1;
            obj.numInputs   = size(inputSpans, 1);
            obj.inputSpans  = inputSpans;

            for i=1:obj.numInputs
                obj.span(i) = obj.inputSpans(i,2) - obj.inputSpans(i,1);                
                %obj.bias1=0.8326/((d_max/(sqrt(2*obj.numRbfs))))^obj.rbfsSpreadDivisor;% use 0.1 for 3 inputs
            end
            placeRbfsGrid(obj);
            obj.LW          = zeros(numRbfs,1) + 0.0;            
            obj.biasOutput  = 0.2025; % bias the output e.g. increase expectations, start with a positive/negative version of the world
            obj.Ntrains     = 0;
        end

        %TRAIN med delta rule
        function obj=train(obj, input, target)       
            obj.Ntrains=obj.Ntrains+1;
            
            if obj.adjBias == 1;
                obj.biasOutput=0.2025;%(obj.biasOutput*(obj.Ntrains-1)+target)/obj.Ntrains;
            end
            output = sim(obj, input);
            error  = target - output;
            updateLW(obj, input, error);
        end

        function updateLW(obj, input, error)
            calcRbf(obj, input);
            correction = obj.rbfOutput * error * obj.lr; 
            obj.LW     = obj.LW + correction;
        end

        function output=sim(obj, input)
            calcRbf(obj,input);
            LWRbf = mtimes(obj.LW', obj.rbfOutput);
            if size(LWRbf) > 1
                pause
            end
            %output = purelin(plus_);
            output=LWRbf+obj.biasOutput;%plus(LWRbf,obj.biasOutput);
        end

        function obj=calcRbf(obj, input)
            obj.rbfOutput=[];
  
            for i=1:size(input,2) % each column  
                my_x=input(:,i)';
                inputRepmat=my_x(ones(1,obj.numRbfs), :);
                obj.rbfOutput(:,i)=exp(-(obj.spread.*sum(bsxfun(@minus,obj.IW,inputRepmat).^2,2).^0.5).^2);
            end
        end


        function placeRBFsRand(obj)
            obj.IW(:,i)=rand(obj.numRbfs,1).*obj.span(i)+obj.inputSpans(i,1);
        end


        function obj=placeRbfsGrid(obj)
            N=obj.numRbfs;%-5;
            if obj.numInputs == 1
                obj.IW = linspace(obj.inputSpans(1,1),obj.inputSpans(1,2),N);
            elseif obj.numInputs == 2
                H1 = linspace(obj.inputSpans(1,1)-0.2,obj.inputSpans(1,2)+0.2,sqrt(N));
                H2 = linspace(obj.inputSpans(2,1)-0.2,obj.inputSpans(2,2)+0.2,sqrt(N));
                [X,Y] = meshgrid(H1,H2);
                AC = reshape(X,1,N);
                ST = reshape(Y,1,N);
                obj.IW = [AC' ST'];
            elseif obj.numInputs == 3
                cr=obj.cubicRoot();
                H1 = linspace(obj.inputSpans(1,1),obj.inputSpans(1,2),cr);
                H2 = linspace(obj.inputSpans(2,1),obj.inputSpans(2,2),cr);
                H3 = linspace(obj.inputSpans(3,1),obj.inputSpans(3,2),cr);
                [X,Y,Z] = meshgrid(H1,H2,H3);
                N
                AC  = reshape(X,1,N);
                ST  = reshape(Y,1,N);
                ST2 = reshape(Z,1,N);
                obj.IW = [AC' ST' ST2'];
            end

        end % end placeRbfsGrid


        function out=cubicRoot(obj)
            out=0;
            for i=1:10 %1000 rbfs maximum
                if i*i*i == obj.numRbfs
                    out = i;
                end
            end
        end

        function setEpochsAndGoal(obj, numEpochs, goal)
            obj.maxEpochs = numEpochs;
            obj.goal = goal;
        end

        function setSpread(obj, spread)
            obj.spread = spread;
        end

        function setLearningRate(obj, lr)
            obj.lr = lr;
        end
        
        function setadjBias(obj, adjBias)
            obj.adjBias = adjBias;
        end
        
        function out = getAdjBias(obj)
            out = obj.adjBias;
        end
    end % end methods
end % end class


