
classdef ReinforcementLearningModel <hgsetget

    properties
        
        dr      %discountRate
        ss      %stepSize
        Q       % Q
        Qprim   % Qprim
        agent
        etType = 'replacing' %'accumulating' or 'replacing'
        omega
        RLalg = 'sarsa';
    end
    properties (SetAccess='private')
        useEt = false;
    end

    methods
        function obj = ReinforcementLearningModel()
            obj.dr = 0.9;
            obj.ss = 0.4;
        end

        function obj = init(obj, agent)
            obj.agent = agent;
        end
        
        %latest action, the new state, the reward recieved
        function omegaTDQL = calcTDerror(obj, aPrim, sPrim, reward, a, s)
            if obj.agent.numTrainings == 0 || obj.agent.numTrainings == 1
                obj.Q = reward*0.01; %if too close to reward it never trains, if too far learning rate needs to be high at first and then lower
            else
                obj.Q = obj.agent.sim([a; s]);
            end
            
            % CALC QMAX AND BEST ACTION
            % Qmax is the hightest Qvalue ie neuralNetwork output
            % if sarsa take   Q(aPrim, sPrim)
            if strcmp(obj.RLalg, 'sarsa')
                obj.Qprim = obj.agent.sim([aPrim; sPrim]);
            else
                bestAction = obj.agent.getBestAction(sPrim);
                obj.Qprim       = obj.agent.sim([bestAction; sPrim]);
            end
            
            % if Q learnign take Qmax
            obj.omega = reward + (obj.dr * obj.Qprim) - obj.Q;
            
            if obj.omega > 100
                %disp(['LR too high TDerror = ' num2str(omegavalue)])
                omegaTDQL = -Inf;
            else
                omegaTDQL = obj.omega;
            end
        end
        function QnewTrace=calcQnewTrace(obj, action, state, omega, trace)            
            QoldTrace=obj.calcQ(action, state);
            QnewTrace = QoldTrace + obj.ss * omega * trace;

        end
        
         function Qnew=calcQnew(obj, action, state, omega)
            obj.Q=obj.calcQ(action, state);
            Qnew = obj.Q + obj.ss * omega;
        end
        
        function omega=getOmega(obj)
            omega=obj.omega;
        end
        
        function Q=calcQ(obj, action, state) 
            Q=obj.agent.sim([action; state']);
        end

        % getters
        function out=getUseEt(obj)
            out = obj.useEt;
        end
        
        % setters
        function setStepSize(obj, ss)
            obj.ss = ss;
        end
        
        function setDiscountRate(obj, dr)
            obj.dr = dr;
        end
        
        function setUseEt(obj, boolean)
            obj.useEt = boolean;
        end
        
        % 'sarsa' or 'Qlearning'
        function setRLalgorithm(obj, name)
            obj.RLalg = name;
        end
        
    end % end methods
end % end class


