% tar: ett state, en action, starttid, sluttid, funktionstyp.
% returnerar: funktionsv?rde, tid, diff, numStates, numActions
function [S]=E(varargin)


if length(varargin)<2
    S.M=varargin{1};
    numInputs = 2;
elseif length(varargin)<3
    S.M=varargin{1};
    numInputs = varargin{2};
elseif length(varargin)<4
    S.M=varargin{1};
    numInputs = varargin{2};
    S.S=varargin{3}; % startv?rde f?r best?ndet
elseif length(varargin)<5
    S=varargin{1};
    A=varargin{2};
    dt=varargin{4};
end


if strcmp('static',S.M)           % static
    if length(varargin)<=3
        S.M='static';
        S.Dim=1;
        S.NA=1; % num actions
        S.Description='Gives a time-INdependent function with a single optima at 0.5';
        %S.S=[10];
        S.P(1)=1;
        S.P(2)=1;
        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(1).sym='r';
        S.text(2).sym='K';
        S.t=0;
        S.dS=0;
        S.S0=0;
        S.AgentInputRanges       =  [0 1.0; 0 S.P(2)];
    else
        S.S0=S.S;
        S.dS=S.P(1)*(1-A)*A*S.P(2);
        S.S=S.dS;
    end
elseif strcmp('logistic',S.M)
    
    if length(varargin)<=3
        S.M='logistic';
        S.Dim=1;
        S.NA=1;
        S.Description='Gives a time-dependent function with a single optima at K * 0.5';
        %S.S=[1]; % startv?rde f?r best?ndet
        S.P(1)=1; % should be 1 growth rate (f?r?ndras)
        S.P(2)=1; % carrying cap
        S.P(3)=0.; %immigtation rate
        %S.P(6)=0.55;%5.02; % because of immigration
        %S.P(7)=0.45;
        S.P(8)=1; % ORIGINAL GROWTH RATE

        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(6).name='Target State';
        S.text(7).name='Best Action';

        S.text(1).sym='r';
        S.text(2).sym='K';
        S.text(3).name='Immigration rate';
        S.text(3).sym='I';
        S.t=0;
        S.dS=0;
        S.S0=S.S;
        
        
        if numInputs == 3
            S.AgentInputRanges       =  [0 1.0; 0 S.P(1); 0 S.P(2)]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        else
            S.AgentInputRanges       =  [0 1.0; 0 S.P(2)];
        end
        
        
        
    else
        S.S0=S.S;
        if dt>1
            for i=1:floor(dt)
                S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S*A+S.P(3));
                S.S=max(0,S.S+S.dS);
            end
        else
            S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S*A+S.P(3))*dt;
            S.S=max(0,S.S+S.dS);
            S.t=S.t+dt;
        end
        S.t=S.t+dt;
    end
elseif strcmp('logistic+rand',S.M)           % static
    if length(varargin)<=2
        S.M='logistic+rand';
        S.Dim=1;
        S.NA=1;
        S.Description='Gives a time-dependent function with a single optima at K * 0.5 and r=rand()';
        S.S=[10];
        S.P(1)=1;
        S.P(2)=10;
        S.P(3)=0.01;
        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(1).sym='r';
        S.text(2).sym='K';
        S.text(3).name='Immigration rate';
        S.text(3).sym='I';
        S.t=0;
        S.dS=0;
        S.S0=S.S;
    else
        S.S0=S.S;
        if dt>1
            for i=1:floor(dt)
                S.dS=(S.P(1)*(1+0.2*(rand-0.5))*(1-S.S/S.P(2))*S.S-S.S*A+S.P(3));
                S.S=max(0,S.S+S.dS);
            end
        else
            S.dS=(S.P(1)*rand*(1-S.S/S.P(2))*S.S-S.S*A+S.P(3))*dt;
            S.S=max(0,S.S+S.dS);
            S.t=S.t+dt;
        end
        S.t=S.t+dt;
    end
    
elseif strcmp('hysteresis',S.M)    % dynamic hysteresis k=q
    if length(varargin)<=3
        S.M='hysteresis';
        S.Dim=1;
        S.NA=1;
        S.Description='Gives a time-dependent function with a single optima at K * 0.5 adn a threshold value at X';
        %S.S=[10];
        S.P(1)=1.47840000; % should be 1.47840000;
        S.P(2)=1.0240;
        S.P(3)=0.0;
        S.P(4)=0.05;
        S.P(5)=0.13;
        S.P(6)=0.55;% draw state toward this value
        S.P(7)=0.45;
        S.P(8)=1.47840000; % ORIGINAL GROWTHRATE
        
        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(3).name='Immigration rate';
        S.text(4).name='Half saturation constant';
        S.text(5).name='corr';
        S.text(6).name='Target State';
        S.text(7).name='Best Action';
        
        S.text(1).sym='r';
        S.text(2).sym='K';
        S.text(3).sym='I';
        S.text(4).sym='k';
        S.t=0;
        S.dS=0;
        S.S0=0;
        
        S.S0=S.S;
        
        if numInputs == 3
            S.AgentInputRanges       =  [0 1.0; 0 S.P(1); 0 S.P(2)]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        else
            S.AgentInputRanges       =  [0 1.0; 0 S.P(2)];
        end
        
        
    else
        S.S0=S.S;
        if dt>1
            for i=1:floor(dt)
                S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*A*corr+S.P(3));
                S.S=max(0,S.S+S.dS);
            end
        else
            S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*S.P(5)-A*S.S+S.P(3))*dt;
            S.S=max(0,S.S+S.dS);
            S.t=S.t+dt;
        end
        S.t=S.t+dt;
    end
    
    %moving threshold 1
elseif strcmp('hysteresis1',S.M)    % dynamic hysteresis k=q
    if length(varargin)<=3
        S.M='hysteresis1';
        S.Dim=1;
        S.NA=1;
        S.Description='Gives a time-dependent function with a single optima at K * 0.5 adn a threshold value at X';
        %S.S=[10];
        S.P(1)=1.47840000; % should be 1.47840000; moves right threshold
        S.P(2)=0.98;%(1.0240;)
        S.P(3)=0.0;
        S.P(4)=0.0445; %(0.05) 0.055 0.6 flytta l?g tr?skel mot h?ger
        S.P(5)=0.11; % (0.13) flytta h?r tr?skel ?t h?ger genom att minska
        S.P(6)=0.55;% draw state toward this value
        S.P(7)=0.45;
        S.P(8)=1.47840000; % ORIGINAL GROWTH RATE

        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(3).name='Immigration rate';
        S.text(4).name='Half saturation constant';
        S.text(5).name='corr';
        S.text(6).name='Target State';
        S.text(7).name='Best Action';
        
        S.text(1).sym='r';
        S.text(2).sym='K';
        S.text(3).sym='I';
        S.text(4).sym='k';
        S.t=0;
        S.dS=0;
        S.S0=0;
        
        S.S0=S.S;
        
        if numInputs == 3
            S.AgentInputRanges       =  [0 1.0; 0 S.P(1); 0 S.P(2)]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        else
            S.AgentInputRanges       =  [0 1.0; 0 S.P(2)];
        end
        
        
    else
        S.S0=S.S;
        if dt>1
            for i=1:floor(dt)
                S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*A*corr+S.P(3));
                S.S=max(0,S.S+S.dS);
            end
        else
            S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*S.P(5)-A*S.S+S.P(3))*dt;
            S.S=max(0,S.S+S.dS);
            S.t=S.t+dt;
        end
        S.t=S.t+dt;
    end
    
    
    %moving threshold 2
elseif strcmp('hysteresis2',S.M)    % dynamic hysteresis k=q
    if length(varargin)<=3
        S.M='hysteresis2';
        S.Dim=1;
        S.NA=1;
        S.Description='Gives a time-dependent function with a single optima at K * 0.5 adn a threshold value at X';
        %S.S=[10];
        S.P(1)=1.47840000; % should be 1.47840000; moves right threshold
        S.P(2)=0.94;%(1.0240;)
        S.P(3)=0.0;
        S.P(4)=0.0382; %(0.05) 0.055 0.6 flytta l?g tr?skel mot h?ger
        S.P(5)=0.09; % (0.13) flytta h?g tr?skel ?t h?ger genom att minska
        S.P(6)=0.55;% draw state toward this value
        S.P(7)=0.45;
        S.P(8)=1.47840000; % should be 1.47840000; moves right threshold

        S.text(1).name='intrinsic growth rate';
        S.text(2).name='Carrying Capacity';
        S.text(3).name='Immigration rate';
        S.text(4).name='Half saturation constant';
        S.text(5).name='corr';
        S.text(6).name='Target State';
        S.text(7).name='Best Action';
        
        S.text(1).sym='r';
        S.text(2).sym='K';
        S.text(3).sym='I';
        S.text(4).sym='k';
        S.t=0;
        S.dS=0;
        S.S0=0;
        
        S.S0=S.S;
        
        if numInputs == 3
            S.AgentInputRanges       =  [0 1.0; 0 S.P(1); 0 S.P(2)]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        else
            S.AgentInputRanges       =  [0 1.0; 0 S.P(2)];
        end
        
        
    else
        S.S0=S.S;
        if dt>1
            for i=1:floor(dt)
                S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*A*corr+S.P(3));
                S.S=max(0,S.S+S.dS);
            end
        else
            S.dS=(S.P(1)*(1-S.S/S.P(2))*S.S-S.S^2/(S.P(4)^2+S.S^2)*S.P(5)-A*S.S+S.P(3))*dt;
            S.S=max(0,S.S+S.dS);
            S.t=S.t+dt;
        end
        S.t=S.t+dt;
    end
elseif strcmp('spatial',S.M)
    if length(varargin)<3
        
        if length(varargin)==2
            try
                temp=varargin{2};
                S.q=temp(1); % location factor
                S.L=temp(2); % fish size
                S.numAgents = temp(3);
            catch
                disp 'need to assign spatial and life-stage resolution like E('spatial',[3 5])'
            end
        end
        S.Dim=5;
        S.NA=3;
        S.Description='This model provides a size-structured spatially semi-explicit model.';
        S.N=2^S.q;      % location
        S.K=1000*S.N;
        S.d=0.05; S.B=0; S.A=0.9; S.a=1; S.phase=5;
        S.S=zeros(1,S.N*S.L)+1;
        S.loc=repmat(1:S.N,1,S.L);
        S.size=reshape(repmat(1:S.L,S.N,1),1,S.L*S.N);
        MG=zeros(S.N*S.L,S.N*S.L);
        MR=zeros(S.N*S.L,S.N*S.L);
        D=zeros(S.N*S.L,S.N*S.L);
        S.As=1;
        L=S.L;   % temp size
        N=S.N;   % temp location
        S.AgentNN  =  [S.NA 40 1231231];  % F?R DEBUGGINNG parameters for the NN 1. numActions 2. num radbas
        %    S.AgentNN  =  [S.NA (S.NA)*20]  % parameters for the NN 1. numActions 2. num radbas
        if S.L==1
            S.AgentInputRanges      =  [0 1; 1 S.N; 0 S.L; 0 S.N; 0 S.N]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        else
            S.AgentInputRanges      =  [0 1; 1 S.N; 0 S.L; 0 S.L; 0 S.L]; % Input ranges: harvest effort 0-1, location 1-S.N, Size 0-S.L,jon kolla,jon kolla.
        end
        S.AgentsActionStepSizes =  [0.05; 1; 1];                   % actions step size (column vector)
        
        for i=1:L
            mig(i)=1-(L-i)./(0.1+(L-i));
        end
        for l1=1:L
            for l2=1:L
                g(l1,l2)=growthRate(l1,l2,L);
                r(l1,l2)=reproductiveRate(l1,l2,L);
            end
        end
        
        
        for l1=1:L
            for l2=1:L
                MG(1+(l1-1)*N:l1*N,1+(l2-1)*N:l2*N)=diag(zeros(1,N)+g(l1,l2));
                MR(1+(l1-1)*N:l1*N,1+(l2-1)*N:l2*N)=diag(zeros(1,N)+r(l1,l2));%r(l1,l2));%r(l1,l2)
            end
        end
        for n1=1:L
            D(1+(n1-1)*N:n1*N,1+(n1-1)*N:n1*N)=diag(zeros(1,N)-1);
            D(2+(n1-1)*N:n1*N,1+(n1-1)*N:n1*N-1)=D(2+(n1-1)*N:n1*N,1+(n1-1)*N:n1*N-1)+diag(zeros(1,N-1)+0.5);
            D(1+(n1-1)*N:n1*N-1,2+(n1-1)*N:n1*N)=D(1+(n1-1)*N:n1*N-1,2+(n1-1)*N:n1*N)+diag(zeros(1,N-1)+0.5);
            D(N+(n1-1)*N,1+(n1-1)*N)=0.5;
            D(1+(n1-1)*N,N+(n1-1)*N)=0.5;
        end
        S.MG=MG;
        S.MR=MR;
        S.D=D;
    else
        h=A(:,1);
        LL=A(:,2);
        s=A(:,3);
        S.h=zeros(S.numAgents,1);
        S.S0=S.S;
        
        S.As = zeros(S.numAgents, 1);
        S.S=max(0,S.S*(S.MG'+S.MR'+S.D*S.d).*(1-sum(S.S)./S.K));
        % loop f?r varje agent
        for x = 1:S.numAgents
            for i = 1:S.N * S.L %location*size
                %[S.loc(i) A(2)]
                
                if (S.size(i)>s(x)) &(S.loc(i)==round(LL(x)))  % if (S.size(i)>A(3))&(S.loc(i)==round(A(2)))
                    %disp 'test loc'
                    S.S(i)=S.S(i)-S.S(i)*h(x);  % A(1) harvest=h
                    S.As(x)=S.As(x)+S.size(i)*S.S(i)*h(x); %
                    S.h(x)=S.h(x)+S.S(i)*h(x);
                end
            end
            if (S.h(x)~=0)
                S.As(x)=S.As(x)/S.h(x);
            else
                S.As(x)=1;
                %     Warning_Average_Size_As_set_to_1_due_to_div_by_zero_because_harvest_h_was_zero = A'
                
            end
        end % end agent loop
    end
else
    error_in_E_function_no_function_type_matched = 0
end

function out=reproductiveRate(x,y,n)
k=0.5;
if x==1
    out=5*((y-1)/(n-1))^2/(k+((y-1)/(n-1)))^2;
else
    out=0;
end


function out=growthRate(x,y,n)
k=0.1;
if y+1==x
    out=(x/n)/(k+(x/n))-0.3/n*x;
else
    out=0;
end
% BTW: seasonality see the do.m file
%B+(A-B)*((sin(t+phase)+1)/2)^a

function out=mymod(x,y)
out=mod(x,y);
if out==0
    out=y;
end
